use super::{archive, package};
use crate::{db, error::Result};

use std::{
    path::{Path, PathBuf},
    sync::Arc,
};

use futures::StreamExt;
use sea_orm::{ColumnTrait, DbConn, ModelTrait, QueryFilter, QuerySelect};
use tokio::{io::AsyncRead, sync::Mutex};
use uuid::Uuid;

pub const ANY_ARCH: &'static str = "any";
pub const REPOS_DIR: &'static str = "repos";
pub const QUEUE_DIR: &'static str = "queue";

pub struct DistroMgr {
    distro_dir: PathBuf,
    distro_id: i32,
    conn: DbConn,
    lock: Arc<Mutex<()>>,
}

impl DistroMgr {
    pub async fn new<P: AsRef<Path>>(distro_dir: P, distro_id: i32, conn: DbConn) -> Result<Self> {
        if !tokio::fs::try_exists(&distro_dir).await? {
            tokio::fs::create_dir(&distro_dir).await?;
        }

        let repos_dir = distro_dir.as_ref().join(REPOS_DIR);

        if !tokio::fs::try_exists(&repos_dir).await? {
            tokio::fs::create_dir(repos_dir).await?;
        }

        let queue_dir = distro_dir.as_ref().join(QUEUE_DIR);

        if !tokio::fs::try_exists(&queue_dir).await? {
            tokio::fs::create_dir(queue_dir).await?;
        }

        Ok(Self {
            distro_dir: distro_dir.as_ref().to_path_buf(),
            distro_id,
            conn,
            lock: Arc::new(Mutex::new(())),
        })
    }

    /// Generate archive databases for all known architectures in the repository, including the
    /// "any" architecture.
    pub async fn generate_archives_all(&self, repo: &str) -> Result<()> {
        let repo = crate::db::query::repo::by_name(&self.conn, repo).await?;

        if repo.is_none() {
            return Ok(());
        }

        let repo = repo.unwrap();

        let mut archs = repo
            .find_related(crate::db::Package)
            .select_only()
            .column(crate::db::package::Column::Arch)
            .distinct()
            .into_tuple::<String>()
            .stream(&self.conn)
            .await?;

        while let Some(arch) = archs.next().await.transpose()? {
            self.generate_archives(&repo.name, &arch).await?;
        }

        Ok(())
    }

    /// Generate the archive databases for the given repository and architecture.
    pub async fn generate_archives(&self, repo: &str, arch: &str) -> Result<()> {
        let repo = crate::db::query::repo::by_name(&self.conn, repo).await?;

        if repo.is_none() {
            return Ok(());
        }

        let repo = repo.unwrap();

        let [tmp_ar_db_path, tmp_ar_files_path, files_tmp_file_path, desc_tmp_file_path] =
            self.random_file_paths();
        let ar_db = archive::RepoArchiveWriter::open(&tmp_ar_db_path).await?;
        let ar_files = archive::RepoArchiveWriter::open(&tmp_ar_files_path).await?;

        // Query all packages in the repo that have the given architecture or the "any"
        // architecture
        let mut pkgs = repo
            .find_related(crate::db::Package)
            .filter(db::package::Column::Arch.is_in([arch, ANY_ARCH]))
            .stream(&self.conn)
            .await?;

        while let Some(pkg) = pkgs.next().await.transpose()? {
            let mut files_tmp_file = tokio::fs::File::create(&files_tmp_file_path).await?;
            let mut desc_tmp_file = tokio::fs::File::create(&desc_tmp_file_path).await?;

            package::write_files(&self.conn, &mut files_tmp_file, &pkg).await?;
            package::write_desc(&self.conn, &mut desc_tmp_file, &pkg).await?;

            let full_name = format!("{}-{}", pkg.name, pkg.version);

            ar_db
                .add_entry(&full_name, &desc_tmp_file_path, true)
                .await?;
            ar_files
                .add_entry(&full_name, &desc_tmp_file_path, true)
                .await?;
            ar_files
                .add_entry(&full_name, &files_tmp_file_path, false)
                .await?;
        }

        // Cleanup
        ar_db.close().await?;
        ar_files.close().await?;

        let repo_dir = self.distro_dir.join(&repo.name);

        // Move the db archives to their respective places
        tokio::fs::rename(tmp_ar_db_path, repo_dir.join(format!("{}.db.tar.gz", arch))).await?;
        tokio::fs::rename(
            tmp_ar_files_path,
            repo_dir.join(format!("{}.files.tar.gz", arch)),
        )
        .await?;

        // If this fails there's no point in failing the function + if there were no packages in
        // the repo, this fails anyway because the temp file doesn't exist
        let _ = tokio::fs::remove_file(desc_tmp_file_path).await;
        let _ = tokio::fs::remove_file(files_tmp_file_path).await;

        Ok(())
    }

    async fn get_or_create_repo(&self, repo: &str) -> Result<db::repo::Model> {
        let _guard = self.lock.lock().await;

        if let Some(repo) = db::query::repo::by_name(&self.conn, repo).await? {
            Ok(repo)
        } else {
            tokio::fs::create_dir(self.distro_dir.join(repo)).await?;

            Ok(db::query::repo::insert(&self.conn, self.distro_id, repo, None).await?)
        }
    }

    /// Remove the repo with the given name, if it existed
    pub async fn remove_repo(&self, repo: &str) -> Result<bool> {
        let res = db::query::repo::by_name(&self.conn, repo).await?;

        if let Some(repo_entry) = res {
            // Remove repository from database
            repo_entry.delete(&self.conn).await?;

            // Remove files from file system
            tokio::fs::remove_dir_all(self.distro_dir.join(repo)).await?;

            Ok(true)
        } else {
            Ok(false)
        }
    }

    /// Remove all packages from the repository with the given arch.
    pub async fn remove_repo_arch(&self, repo: &str, arch: &str) -> Result<bool> {
        let repo = db::query::repo::by_name(&self.conn, repo).await?;

        if let Some(repo) = repo {
            let mut pkgs = repo
                .find_related(db::Package)
                .filter(db::package::Column::Arch.eq(arch))
                .stream(&self.conn)
                .await?;

            while let Some(pkg) = pkgs.next().await.transpose()? {
                let path = self
                    .distro_dir
                    .join(&repo.name)
                    .join(super::package::filename(&pkg));
                tokio::fs::remove_file(path).await?;

                pkg.delete(&self.conn).await?;
            }

            tokio::fs::remove_file(
                self.distro_dir
                    .join(&repo.name)
                    .join(format!("{}.db.tar.gz", arch)),
            )
            .await?;
            tokio::fs::remove_file(
                self.distro_dir
                    .join(&repo.name)
                    .join(format!("{}.files.tar.gz", arch)),
            )
            .await?;

            // If we removed all "any" packages, we need to resync all databases
            if arch == ANY_ARCH {
                self.generate_archives_all(&repo.name).await?;
            }

            Ok(true)
        } else {
            Ok(false)
        }
    }

    pub async fn remove_pkg(&self, repo: &str, arch: &str, name: &str) -> Result<bool> {
        let repo = db::query::repo::by_name(&self.conn, repo).await?;

        if let Some(repo) = repo {
            let pkg =
                db::query::package::by_fields(&self.conn, repo.id, arch, name, None, None).await?;

            if let Some(pkg) = pkg {
                // Remove package from database & file system
                tokio::fs::remove_file(
                    self.distro_dir
                        .join(&repo.name)
                        .join(super::package::filename(&pkg)),
                )
                .await?;
                pkg.delete(&self.conn).await?;

                if arch == ANY_ARCH {
                    self.generate_archives_all(&repo.name).await?;
                } else {
                    self.generate_archives(&repo.name, arch).await?;
                }

                Ok(true)
            } else {
                Ok(false)
            }
        } else {
            Ok(false)
        }
    }

    pub async fn add_pkg_from_reader<R: AsyncRead + std::marker::Unpin>(
        &self,
        reader: &mut R,
        repo: &str,
    ) -> crate::Result<(String, String, String)> {
        let [tmp_file_path] = self.random_file_paths();
        let mut temp_file = tokio::fs::File::create(&tmp_file_path).await?;

        tokio::io::copy(reader, &mut temp_file).await?;

        let path_clone = tmp_file_path.clone();
        let pkg = tokio::task::spawn_blocking(move || package::Package::open(path_clone))
            .await
            .unwrap()?;

        let repo = self.get_or_create_repo(repo).await?;
        let pkg = db::query::package::insert(&self.conn, repo.id, pkg).await?;

        let queue_path = self.distro_dir.join(QUEUE_DIR).join(pkg.id.to_string());
        tokio::fs::rename(tmp_file_path, queue_path).await?;

        // If the package already exists in the database, we remove it first
        //let res = db::query::package::by_fields(
        //    &self.conn,
        //    repo.id,
        //    &pkg.info.arch,
        //    &pkg.info.name,
        //    None,
        //    None,
        //)
        //.await?;
        //
        //if let Some(entry) = res {
        //    entry.delete(&self.conn).await?;
        //}

        //let dest_pkg_path = repo_dir.join(pkg.file_name());
        //
        //// Insert new package into database
        //let name = pkg.info.name.clone();
        //let version = pkg.info.version.clone();
        //let arch = pkg.info.arch.clone();
        //db::query::package::insert(&self.conn, repo.id, pkg).await?;
        //
        //// Move the package to its final resting place
        //tokio::fs::rename(tmp_file_path, dest_pkg_path).await?;
        //
        // Synchronize archive databases
        //if arch == ANY_ARCH {
        //    self.generate_archives_all(&repo.name).await?;
        //} else {
        //    self.generate_archives(&repo.name, &arch).await?;
        //}

        Ok((pkg.name, pkg.version, pkg.arch))
    }

    ///  Generate a path to a unique file that can be used as a temporary file
    pub fn random_file_paths<const C: usize>(&self) -> [PathBuf; C] {
        std::array::from_fn(|_| {
            let uuid: uuid::fmt::Simple = Uuid::new_v4().into();
            self.distro_dir.join(uuid.to_string())
        })
    }
}
